//----------------------------------------------------------------------------
//	SOURCE:	FFio.c
//	AUTHOR:	Ian Taylor - chilliant.com
//	NOTES:	Flag Factory textual output library
//----------------------------------------------------------------------------

#include "FF.h"

void (*FFputc)(char ch) = NULL;

#if FF_FINAL

#include <conio.h>
#include <stdarg.h>
#if FF_OWC
extern void _putch(char chr);
#endif

#else

#include <stdio.h>

#pragma warning(push)
#pragma warning(disable: 4001 4255 4668 4820)

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>

#pragma warning(pop)

static FILE* FFfp = NULL;

void FFredirect(const char* filename)
{
    if (FFfp)
    {
        fclose(FFfp);
        FFfp = NULL;
    }
    if (filename)
    {
#ifdef __DMC__
        FFfp = fopen(filename, "wt");
#else
        errno_t err;
        err = fopen_s(&FFfp, filename, "wt");
        ASSERT(err == 0);
#endif
    }
}

int FFdebug(const char* file, int line, const char* message)
{
    char buffer[1024];
    sprintf_s(buffer, sizeof(buffer), "%s(%d): %s\n", file, line, message);
    OutputDebugStringA(buffer);
    return IsDebuggerPresent();
}

#endif

void FFnull(char ch)
{
    ASSERT((ch > 1) && (ch < 127));
}

void FFio(char chr)
{
    ASSERT((chr > 1) && (chr < 127));
#if FF_FINAL
	_putch(chr);
#else
    if (FFfp)
    {
        fputc(chr, FFfp);
    }
    else
    {
        char text[2];
        text[0] = chr;
        text[1] = '\0';
        OutputDebugStringA(text);
    }
#endif
}

void FFputs(const char* text)
{
    ASSERT(text);
    while (*text)
    {
        FFputc(*(text++));
    }
}

void FFputz(FFfar text)
{
    ASSERT(text);
    while (*text)
    {
        FFputc(*(text++));
    }
}

void FFputi(s16 value)
{
    if (value < 0)
    {
        FFputc('-');
        value = -value;
		ASSERT(value > 0);
    }
    FFputu((u16)value);
}

void FFputu(u16 value)
{
    u16 tens = value / 10;
    char digit;
    if (tens)
    {
        FFputu(tens);
    }
    digit = (char)(value - tens * 10) + '0';
    ASSERT((digit >= '0') && (digit <= '9'));
    FFputc(digit);
}

void FFputf(fp value)
{
	const s16 mscale = 10000;
    s16 i = FFs16(value);
    fp f = FFabs(value - i);
    s16 m = FFs16(f * mscale + 0.5f);
    ASSERT((value >= -0x7FFF) && (value <= 0x7FFF));
    ASSERT((f >= 0) && (f < 1));
    ASSERT(m <= mscale);
    if (m == mscale)
    {
        if (i < 0)
        {
            i--;
        }
        else
        {
            i++;
        }
        m = 0;
    }
    FFputi(i);
    if (m != 0)
    {
		s16 dscale = mscale;
		char digit;
        FFputc('.');
        do
        {
			ASSERT(dscale > 0);
			dscale /= 10;
            digit = (char)(m / dscale);
            FFputc(digit + '0');
			m -= digit * dscale;
        }
        while (m != 0);
    }
}

void FFput2i(char* first, s16 x, s16 y)
{
    ASSERT(first);
    FFputc(*first);
    FFputi(x);
    FFputc(',');
    FFputi(y);
    *first = ' ';
}

void FFput2f(char* first, fp x, fp y)
{
    ASSERT(first && ((*first == '"') || (*first == ' ')));
    FFputc(*first);
    FFputf(x);
    FFputc(',');
    FFputf(y);
    *first = ' ';
}

static void FFputv(const char* fmt, va_list* vargs)
{
    ASSERT(fmt);
    while (*fmt)
    {
        switch (*fmt)
        {
        case 'A':
            FFputv(" S=\"S\"", vargs);
            break;
        case 'B':
            FFputv(" S=\"V\"", vargs);
            break;
        case 'C':
            FFputv(" fill=\"#S\"", vargs);
            break;
        case 'D':
            FFputv(" S=\"I\"", vargs);
            break;
        case 'E':
            FFputv(" S=\"F\"", vargs);
            break;
        case 'F':
            FFputf((fp)va_arg(*vargs, double));
            break;
        case 'I':
            FFputi(va_arg(*vargs, s16));
            break;
        case 'S':
            FFputs(va_arg(*vargs, const char*));
            break;
        case 'U':
            FFputu(va_arg(*vargs, u16));
            break;
        case 'V':
            FFputv(va_arg(*vargs, const char*), vargs);
            break;
        case 'Z':
            FFputz(va_arg(*vargs, FFfar));
            break;
        default:
            ASSERT((*fmt < 'A') || (*fmt > 'Z'));
            FFputc(*fmt);
            break;
        }
        fmt++;
    }
}

void FFput(const char* fmt, ...)
{
    va_list vargs;
    va_start(vargs, fmt);
    FFputv(fmt, &vargs);
    va_end(vargs);
}

