//----------------------------------------------------------------------------
//	SOURCE:	FFsvg.c
//	AUTHOR:	Ian Taylor - chilliant.com
//	NOTES:	Flag Factory SVG output module
//----------------------------------------------------------------------------

#include "FF.h"

#define FF_DEBUG_CURVE 0

static void SVG_command(void);
static void SVG_context(void);

static const char SVGcolour_rgb[][4] =
{
	"FFF" // Initial colour
#define COLOUR_LIST_RGB(_,r,g,b) , #r #g #b
    COLOUR_LIST(COLOUR_LIST_RGB)
};

static void SVG_colour(u8 arg)
{
    FFparse.fill = SVGcolour_rgb[arg];
}

static void SVG_rectangle_ex(const char* rgb, s16 x0, s16 y0, s16 x1, s16 y1)
{
    ASSERT(x1 > x0);
    ASSERT(y1 > y0);
    FFput("<rectCDDDD/>\n", rgb,
          "x", x0,
          "y", y0,
          "width", x1 - x0,
          "height", y1 - y0);
}

static void SVG_blockh(u8 arg)
{
    const char* rgb = SVGcolour_rgb[arg];
    s16 y0 = FFgeti();
    s16 y1 = FFgeti();
    SVG_rectangle_ex(rgb, -120, y0, 120, y1);
}

static void SVG_blockv(u8 arg)
{
    const char* rgb = SVGcolour_rgb[arg];
    s16 x0 = FFgeti();
    s16 x1 = FFgeti();
    SVG_rectangle_ex(rgb, x0, -120, x1, 120);
}

static void SVG_circle(u8 arg)
{
    s16 cx = FFgeti();
    s16 cy = FFgeti();
    s16 ry = arg ? arg : FFgeti();
    fp rx = ry / FFparse.aspect_ratio;
	ASSERT(rx > 0);
    ASSERT(ry > 0);
    FFput("<ellipseCDDED/>",  FFparse.fill,
          "cx", cx,
          "cy", cy,
          "rx", rx,
          "ry", ry);
}

static void SVG_cross(u8 arg)
{
	const char* rgb = SVGcolour_rgb[arg];
    s16 w = FFgeti();
    s16 x = FFgeti();
    fp fw = w * FFparse.aspect_ratio;
    ASSERT(w > 0);
    SVG_rectangle_ex(rgb, x - w, -120, x + w, 120);
    FFput("<rectCDEDE/>\n", rgb,
          "x", -120,
          "y", -fw,
          "width", 240,
          "height", fw * 2);
}

#if FF_DEBUG_CURVE
static void SVG_curve_debug(const s16 xy[2], const s16 dxy[2])
{
	FFput("<line stroke=\"lime\" stroke-width=\"1px\" x1=\"I\" y1=\"I\" x2=\"I\" y2=\"I\"/>\n",
			xy[0], xy[1], xy[0] + dxy[0], xy[1] + dxy[1]);
}
#endif

static void SVG_curve(u8 arg)
{
#define MAX_N 64
	static char pc[MAX_N];
	static s16 pt[MAX_N][3][2];
	u16 i, n = 1;
	s8 curve = CmdCurve_linear;
	u16 triple;
	pc[0] = 'M';
	pt[0][0][0] = FFgeti();
	pt[0][0][1] = FFgeti();
	ASSERT(*FFparse.ptr != CmdCurve_end);
	do
	{
		switch (*FFparse.ptr)
		{
		case CmdCurve_linear:
		case CmdCurve_smooth:
		case CmdCurve_cubic:
			ASSERT(curve != *FFparse.ptr);
			curve = *(FFparse.ptr++);
			break;
		default:
			ASSERT(n < MAX_N);
			switch (curve)
			{
			case CmdCurve_linear:
				pc[n] = 'L';
				pt[n][0][0] = FFgeti();
				pt[n][0][1] = FFgeti();
				break;
			case CmdCurve_smooth:
				ASSERT(n > 0);
				ASSERT(pc[n-1] == 'C');
				pc[n] = 'C';
				pt[n][2][0] = FFgeti();
				pt[n][2][1] = FFgeti();
				pt[n][1][0] = -pt[n-1][2][0];
				pt[n][1][1] = -pt[n-1][2][1];
				pt[n][0][0] = FFgeti();
				pt[n][0][1] = FFgeti();
#if FF_DEBUG_CURVE
				SVG_curve_debug(pt[n-1][0], pt[n][1]);
				SVG_curve_debug(pt[n][0], pt[n][2]);
#endif
				break;
			case CmdCurve_cubic:
				pc[n] = 'C';
				pt[n][1][0] = FFgeti();
				triple = FFgetu();
				triple = (triple << 8) | FFgetu();
				pt[n][1][1] = (triple >> 10) - 15;
				pt[n][2][0] = ((triple >> 5) & 31) - 15;
				pt[n][2][1] = (triple & 31) - 15;
				pt[n][0][0] = FFgeti();
				pt[n][0][1] = FFgeti();
#if FF_DEBUG_CURVE
				SVG_curve_debug(pt[n-1][0], pt[n][1]);
				SVG_curve_debug(pt[n][0], pt[n][2]);
#endif
				break;
			}
			n++;
			break;
		}
	}
	while (*FFparse.ptr != CmdCurve_end);
	FFparse.ptr++;
    FFput("<pathC d=\"", FFparse.fill);
	for (i = 0; i < n; ++i)
	{
		char ch = pc[i];
		if (ch == 'C')
		{
			ASSERT(i > 0);
			FFput2i(&ch, pt[i-1][0][0] + pt[i][1][0], pt[i-1][0][1] + pt[i][1][1]);
			FFput2i(&ch, pt[i][0][0] + pt[i][2][0], pt[i][0][1] + pt[i][2][1]);
		}
		FFput2i(&ch, pt[i][0][0], pt[i][0][1]);
	}
	if (arg)
	{
		s16 mx = pt[0][0][0] * 2;
		for (i = n - 1; i > 0; --i)
		{
			char ch = pc[i];
			if (ch == 'C')
			{
				ASSERT(i > 0);
				FFput2i(&ch, mx - pt[i][0][0] - pt[i][2][0], pt[i][0][1] + pt[i][2][1]);
				FFput2i(&ch, mx - pt[i-1][0][0] - pt[i][1][0], pt[i-1][0][1] + pt[i][1][1]);
			}
			FFput2i(&ch, mx - pt[i-1][0][0], pt[i-1][0][1]);
		}
	}
	FFputs(" Z\"/>");
#undef MAX_N
}

static void SVG_diagonal_ex(s16 w, s16 mirror)
{
    char first = '"';
    FFput("<polygonC points=",  FFparse.fill);
    FFput2i(&first, mirror * -120, -120);
    FFput2i(&first, mirror * (-120 + w), -120);
    FFput2i(&first, mirror * +120, +120 - w);
    FFput2i(&first, mirror * +120, +120);
    FFput2i(&first, mirror * (+120 - w), +120);
    FFput2i(&first, mirror * -120, -120 + w);
    FFputs("\"/>\n");
}

static void SVG_diagonal(u8 arg)
{
    s16 w = FFgetu();
    if (arg & 1)
    {
        SVG_diagonal_ex(w, -1);
    }
    if (arg & 2)
    {
        SVG_diagonal_ex(w, +1);
    }
}

static void SVG_flood(u8 arg)
{
    const char* rgb = SVGcolour_rgb[arg];
    SVG_rectangle_ex(rgb, -120, -120, +120, +120);
}

static void SVG_forr(u8 arg)
{
	u16 i, n = arg ? arg : FFgetu();
	fp da = 360.f / n;
	fp aspect_ratio = FFparse.aspect_ratio;
	const s8* stack = FFparse.ptr;
	ASSERT(n > 1);
	for (i = 0; i < n; ++i)
	{
		FFput("<gB>\n", "transform", "scale(F,1) rotate(F)", 1.f / aspect_ratio, da * i);
		FFparse.aspect_ratio = 1;
		FFparse.ptr = stack;
		SVG_command();
		FFput("</g>\n");
	}
	FFparse.aspect_ratio = aspect_ratio;
}

static void SVG_fors_ex(fp sx, fp sy)
{
	FFput("<gB>\n ", "transform", "scale(F,F)", sx, sy);
	SVG_command();
	FFput("</g>\n");
}

static void SVG_fors(u8 arg)
{
	fp sx = (fp)arg - 1;
	fp sy = FFgetf() / 60.f;
	if ((sx < 0) || (sy < 0))
	{
		const s8* stack = FFparse.ptr;
		SVG_fors_ex(FFabs(sx), FFabs(sy));
		FFparse.ptr = stack;
	}
	SVG_fors_ex(sx, sy);
}

static void SVG_fort(u8 arg)
{
	u8 i;
	s16 dx = FFgeti();
	s16 dy = FFgeti();
	const s8* stack = FFparse.ptr;
	ASSERT(arg > 1);
	SVG_command();
	for (i = 1; i < arg; ++i)
	{
		FFput("<gB>\n ", "transform", "translate(I,I)", dx * i, dy * i);
		FFparse.ptr = stack;
		SVG_command();
		FFput("</g>\n");
	}
}

static void SVG_rectangle(void)
{
    s16 x0 = FFgeti();
    s16 y0 = FFgeti();
    s16 x1 = FFgeti();
    s16 y1 = FFgeti();
    SVG_rectangle_ex(FFparse.fill, x0, y0, x1, y1);
}

static void SVG_star_ex(u16 n, fp r1, fp r2, fp rot)
{
    fp x = FFgetf();
    fp y = FFgetf();
    char first = '"';
    u16 i;
    if (r2 == 0)
    {
        r2 = r1 * 0.385f; // WIBBLE
    }
    ASSERT(n >= 3);
    ASSERT(r1 > 0);
    ASSERT(r2 > 0);
    FFput("<polygonC points=", FFparse.fill);
    for (i = 0; i < (n * 2); i++)
    {
		fp angle = 3.1415927f * i / n + rot;
		fp ry = r1;
		fp rx = ry / FFparse.aspect_ratio;
		r1 = r2;
		r2 = ry;
        FFput2f(&first, x + FFsin(angle) * rx, y + FFcos(angle) * ry);
    }
    FFputs("\"/>\n");
}

static void SVG_star(u8 arg)
{
    u16 n = arg + 4;
    fp r1 = FFgetf();
    fp r2 = FFgetf();
    fp rot = FFgetf() * 3.1415927f / 120.f;
	SVG_star_ex(n, r1, r2, rot);
}

static void SVG_star5(u8 arg)
{
    fp r;
	if (arg)
	{
		// Small star
		r = (fp)arg + 1;
	}
	else
	{
		// Large star
		r = FFgetf();
		ASSERT(r > 16);
	}
	SVG_star_ex(5, r, 0, 0);
}

static void SVG_stripes_ex(int horz, u16 colours)
{
    u16 stripes;
    const char* rgb[4];
    u16 i;
    fp x, y, z, dx, dy, dz, sx, sy;
	u8 arg[6];
	u8 got = 0;
    ASSERT((colours >= 2) && (colours <= 4));
	do
	{
		ASSERT(got < sizeof(arg));
		arg[got++] = (u8)*FFparse.ptr & 0x0F;
		arg[got++] = (u8)*FFparse.ptr >> 4;
		FFparse.ptr++;
	}
	while (got <= colours);
    stripes = arg[0];
    ASSERT(stripes >= 2);
    for (i = 0; i < colours; i++)
    {
        rgb[i] = SVGcolour_rgb[arg[i + 1]];
    }
	x = -120.f;
    z = 241.f / stripes;
	dz = (240.f - z) / (stripes - 1);
	ASSERT(dz < z);
    if (horz)
    {
        y = 120.f - z;
        dx = 0.f;
        dy = -dz;
        sx = 240.f;
        sy = z;
    }
    else
    {
        y = -120.f;
        dx = dz;
        dy = 0.f;
        sx = z;
        sy = 240.f;
    }
    for (i = 0; i < stripes; i++)
    {
        FFput("<rectCEEEE/>\n", rgb[i % colours],
              "x", x,
              "y", y,
              "width", sx,
              "height", sy);
        x += dx;
        y += dy;
    }
}

static void SVG_stripesh(u8 arg)
{
	SVG_stripes_ex(1, arg);
}

static void SVG_stripesv(u8 arg)
{
	SVG_stripes_ex(0, arg);
}

static void SVG_subflag(const char* flag, int quarter)
{
    if (FFputc != FFnull)
    {
        FFparse_s stack = FFparse;
		ASSERT(FFputc == FFio);
		if (quarter)
		{
			FFput("<gA>\n", "transform", "matrix(0.5,0,0,0.5,-60,60)");
		}
		ASSERT(flag && flag[0] && flag[1] && !flag[2]);
		FFparse.ptr = FFdata;
		FFparse.name = FFname;
		FFputc = FFnull;
		ASSERT(*FFparse.ptr);
		for (;;)
		{
			// Iterate around each flag entry in the table
			char next[3];
			const char* name = FFdecode(next);
			if (*(const u16*)flag == *(const u16*)next)
			{
				break;
			}
			FFsvg(name);
		}
		FFputc = FFio;
        SVG_context();
        FFparse = stack;
		if (quarter)
		{
	        FFput("</g>\n");
		}
    }
}

#if !FF_TINY
static void SVG_alias(void)
{
	char subflag[3];
	subflag[0] = *(FFparse.ptr++);
	subflag[1] = *(FFparse.ptr++);
	subflag[2] = '\0';
	SVG_subflag(subflag, 0);
}
#endif

static void SVG_jack(void)
{
	SVG_subflag("gb", 1);
}

static void SVG_triangle(void)
{
    s16 x[3], y[3];
	s16 sx, sy;
    char first = '"';
    u16 i;
    x[0] = FFgeti();
    y[1] = FFgeti();
    x[1] = FFgeti();
    y[0] = FFgeti();
	sx = x[1] - x[0];
	ASSERT(sx != 0);
	sy = y[1] - y[0];
	ASSERT(sy != 0);
    i = (sx < 0) ^ (sy < 0);
    x[2] = x[1 - i];
    y[2] = y[i];
    FFput("<polygonC points=", FFparse.fill);
    for (i = 0; i < 3; i++)
    {
        FFput2i(&first, x[i], y[i]);
    }
    FFputs("\"/>\n");
}

static void SVG_viewbox(void)
{
    s16 x0 = FFgeti();
    s16 y0 = FFgeti();
    s16 x1 = FFgeti();
    s16 y1 = FFgeti();
    fp sx = (x1 - x0) / 240.f;
    fp sy = (y1 - y0) / 240.f;
    fp tx = (x0 + x1) * 0.5f;
    fp ty = (y0 + y1) * 0.5f;
    FFput("</g><gB>\n",
          "transform", "matrix(F,0,0,F,F,F)",
          sx, sy, tx, ty);
	FFparse.aspect_ratio *= sx / sy;
}

static void SVG_Ext(u8 arg)
{
	typedef void (*JUMP)(void);
	static JUMP const jump[] =
	{
		NULL // CmdExt_End
#define CMDEXT_LIST_JUMP(x) , SVG_##x
        CMDEXT_LIST(CMDEXT_LIST_JUMP)
	};
	ASSERT(arg < (sizeof(jump) / sizeof(*jump)));
	ASSERT(jump[arg]);
	jump[arg]();
}

static void SVG_command_ex(u8 cmd)
{
	typedef void (*JUMP)(u8 arg);
	static JUMP const jump[] =
	{
		SVG_Ext // Cmd_ext
#define CMD_LIST_JUMP(x) , SVG_##x
        CMD_LIST(CMD_LIST_JUMP)
	};
	ASSERT((sizeof(jump) / sizeof(*jump)) == 16);
	jump[CMD_GET_CMD(cmd)](CMD_GET_ARG(cmd));
}

static void SVG_command(void)
{
	SVG_command_ex(FFgetu());
}

static void SVG_context(void)
{
    u8 cmd = FFgetu();
    while (cmd != Cmd_End)
    {
		SVG_command_ex(cmd);
        cmd = FFgetu();
    }
}

const char* FFdecode(char flag[3])
{
	static const u8 ar_table[] =
	{
		#define ASPECT_RATIO_TABLE(x,y) x,
		ASPECT_RATIO(ASPECT_RATIO_TABLE)
	};
	u16 triple = FFgetu();
	const char* name = FFparse.name;
	triple = (triple << 8) | FFgetu();
	flag[0] = ((triple >> 11) & 31) + 'a' - 1;
	ASSERT((flag[0] >= 'a') && (flag[0] <= 'z'));
	flag[1] = ((triple >> 6) & 31) + 'a' - 1;
	ASSERT((flag[1] >= 'a') && (flag[1] <= 'z'));
	flag[2] = '\0';
	ASSERT(flag[0] == name[0]);
	ASSERT(flag[1] == name[1]);
	triple &= 63;
	ASSERT(AspectRatio_Count <= 64);
	ASSERT(triple < sizeof(ar_table) - 1);
	ASSERT(ar_table[triple]);
	ASSERT(ar_table[triple+1]);
    FFparse.aspect_ratio = (fp)ar_table[triple] / ar_table[triple+1];
	ASSERT(FFparse.name && *FFparse.name);
	while (*(FFparse.name++)) {}
	return name;
}

void FFsvg(const char* name)
{
    FFput("<?xmlAA?>\n<S><svgAB>\n",
		  "version", "1.0",
		  "encoding", "UTF-8",
		  "!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1 Tiny//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11-tiny.dtd\"",
          "xmlns", "http://www.w3.org/2000/svg",
          "viewBox", "0 0 F 240", FFparse.aspect_ratio * 240);
    FFput("<title>S - chilliant.com</title>\n<gBA><g>\n",
		  name,
          "transform", "matrix(F,0,0,-1,F,120)", FFparse.aspect_ratio, FFparse.aspect_ratio * 120,
          "stroke", "none");
    FFparse.fill = SVGcolour_rgb[0];
    SVG_context();
    FFput("\n</g></g></svg>\n");
}
